[<prev] [next>] [<thread-prev] [thread-next>] [day] [month] [year] [list]
Message-Id: <20220220134813.3411982-7-memxor@gmail.com>
Date: Sun, 20 Feb 2022 19:18:04 +0530
From: Kumar Kartikeya Dwivedi <memxor@...il.com>
To: bpf@...r.kernel.org
Cc: Alexei Starovoitov <ast@...nel.org>,
Andrii Nakryiko <andrii@...nel.org>,
Daniel Borkmann <daniel@...earbox.net>,
Toke Høiland-Jørgensen <toke@...hat.com>,
Jesper Dangaard Brouer <hawk@...nel.org>,
netfilter-devel@...r.kernel.org, netdev@...r.kernel.org
Subject: [PATCH bpf-next v1 06/15] bpf: Allow storing __user PTR_TO_BTF_ID in map
Recently, verifier gained __user annotation support [0] where it
prevents BPF program from normally derefering user memory pointer in the
kernel, and instead requires use of bpf_probe_read_user. We can allow
the user to also store these pointers in BPF maps, with the logic that
whenever user loads it from the BPF map, it gets marked as MEM_USER.
[0]: https://lore.kernel.org/bpf/20220127154555.650886-1-yhs@fb.com
Signed-off-by: Kumar Kartikeya Dwivedi <memxor@...il.com>
---
include/linux/bpf.h | 1 +
kernel/bpf/btf.c | 20 +++++++++++++++-----
kernel/bpf/verifier.c | 21 +++++++++++++++------
3 files changed, 31 insertions(+), 11 deletions(-)
diff --git a/include/linux/bpf.h b/include/linux/bpf.h
index 843c8c01cf9d..37ca92f4c7b7 100644
--- a/include/linux/bpf.h
+++ b/include/linux/bpf.h
@@ -163,6 +163,7 @@ enum {
enum {
BPF_MAP_VALUE_OFF_F_REF = (1U << 0),
BPF_MAP_VALUE_OFF_F_PERCPU = (1U << 1),
+ BPF_MAP_VALUE_OFF_F_USER = (1U << 2),
};
struct bpf_map_value_off_desc {
diff --git a/kernel/bpf/verifier.c b/kernel/bpf/verifier.c
index 00d6ab49033d..28da858bb921 100644
--- a/kernel/bpf/verifier.c
+++ b/kernel/bpf/verifier.c
@@ -3482,7 +3482,11 @@ static int map_ptr_to_btf_id_match_type(struct bpf_verifier_env *env,
enum bpf_reg_type reg_type;
const char *reg_name = "";
- if (off_desc->flags & BPF_MAP_VALUE_OFF_F_PERCPU) {
+ if (off_desc->flags & BPF_MAP_VALUE_OFF_F_USER) {
+ if (reg->type != (PTR_TO_BTF_ID | MEM_USER) &&
+ reg->type != (PTR_TO_BTF_ID | PTR_MAYBE_NULL | MEM_USER))
+ goto end;
+ } else if (off_desc->flags & BPF_MAP_VALUE_OFF_F_PERCPU) {
if (reg->type != PTR_TO_PERCPU_BTF_ID &&
reg->type != (PTR_TO_PERCPU_BTF_ID | PTR_MAYBE_NULL))
goto end;
@@ -3536,7 +3540,9 @@ static int map_ptr_to_btf_id_match_type(struct bpf_verifier_env *env,
return 0;
end:
- if (off_desc->flags & BPF_MAP_VALUE_OFF_F_PERCPU)
+ if (off_desc->flags & BPF_MAP_VALUE_OFF_F_USER)
+ reg_type = PTR_TO_BTF_ID | PTR_MAYBE_NULL | MEM_USER;
+ else if (off_desc->flags & BPF_MAP_VALUE_OFF_F_PERCPU)
reg_type = PTR_TO_PERCPU_BTF_ID | PTR_MAYBE_NULL;
else
reg_type = PTR_TO_BTF_ID | PTR_MAYBE_NULL;
@@ -3556,14 +3562,14 @@ static int check_map_ptr_to_btf_id(struct bpf_verifier_env *env, u32 regno, int
struct bpf_reg_state *atomic_load_reg)
{
struct bpf_reg_state *reg = reg_state(env, regno), *val_reg;
+ bool ref_ptr = false, percpu_ptr = false, user_ptr = false;
struct bpf_insn *insn = &env->prog->insnsi[insn_idx];
enum bpf_reg_type reg_type = PTR_TO_BTF_ID;
- bool ref_ptr = false, percpu_ptr = false;
struct bpf_map_value_off_desc *off_desc;
int insn_class = BPF_CLASS(insn->code);
+ int ret, reg_flags = PTR_MAYBE_NULL;
struct bpf_map *map = reg->map_ptr;
u32 ref_obj_id = 0;
- int ret;
/* Things we already checked for in check_map_access:
* - Reject cases where variable offset may touch BTF ID pointer
@@ -3590,8 +3596,11 @@ static int check_map_ptr_to_btf_id(struct bpf_verifier_env *env, u32 regno, int
ref_ptr = off_desc->flags & BPF_MAP_VALUE_OFF_F_REF;
percpu_ptr = off_desc->flags & BPF_MAP_VALUE_OFF_F_PERCPU;
+ user_ptr = off_desc->flags & BPF_MAP_VALUE_OFF_F_USER;
if (percpu_ptr)
reg_type = PTR_TO_PERCPU_BTF_ID;
+ else if (user_ptr)
+ reg_flags |= MEM_USER;
if (is_xchg_insn(insn)) {
/* We do checks and updates during register fill call for fetch case */
@@ -3623,7 +3632,7 @@ static int check_map_ptr_to_btf_id(struct bpf_verifier_env *env, u32 regno, int
}
/* val_reg might be NULL at this point */
mark_btf_ld_reg(env, cur_regs(env), value_regno, reg_type, off_desc->btf,
- off_desc->btf_id, PTR_MAYBE_NULL);
+ off_desc->btf_id, reg_flags);
/* __mark_ptr_or_null_regs needs ref_obj_id == id to clear
* reference state for ptr == NULL branch.
*/
@@ -3641,7 +3650,7 @@ static int check_map_ptr_to_btf_id(struct bpf_verifier_env *env, u32 regno, int
* value from map as PTR_TO_BTF_ID, with the correct type.
*/
mark_btf_ld_reg(env, cur_regs(env), value_regno, reg_type, off_desc->btf,
- off_desc->btf_id, PTR_MAYBE_NULL);
+ off_desc->btf_id, reg_flags);
val_reg->id = ++env->id_gen;
} else if (insn_class == BPF_STX) {
if (WARN_ON_ONCE(value_regno < 0))
diff --git a/kernel/bpf/btf.c b/kernel/bpf/btf.c
index eb57584ee0a8..bafceae90c32 100644
--- a/kernel/bpf/btf.c
+++ b/kernel/bpf/btf.c
@@ -3146,7 +3146,7 @@ static s32 btf_find_by_name_kind_all(const char *name, u32 kind, struct btf **bt
static int btf_find_field_kptr(const struct btf *btf, const struct btf_type *t,
u32 off, int sz, void *data)
{
- bool btf_id_tag = false, ref_tag = false, percpu_tag = false;
+ bool btf_id_tag = false, ref_tag = false, percpu_tag = false, user_tag = false;
struct bpf_map_value_off *tab;
struct bpf_map *map = data;
int nr_off, ret, flags = 0;
@@ -3181,6 +3181,13 @@ static int btf_find_field_kptr(const struct btf *btf, const struct btf_type *t,
goto end;
}
percpu_tag = true;
+ } else if (!strcmp("kernel.bpf.user", __btf_name_by_offset(btf, t->name_off))) {
+ /* repeated tag */
+ if (user_tag) {
+ ret = -EINVAL;
+ goto end;
+ }
+ user_tag = true;
} else if (!strncmp("kernel.", __btf_name_by_offset(btf, t->name_off),
sizeof("kernel.") - 1)) {
/* TODO: Should we reject these when loading BTF? */
@@ -3192,15 +3199,16 @@ static int btf_find_field_kptr(const struct btf *btf, const struct btf_type *t,
t = btf_type_by_id(btf, t->type);
}
if (!btf_id_tag) {
- /* 'ref' or 'percpu' tag must be specified together with 'btf_id' tag */
- if (ref_tag || percpu_tag) {
+ /* 'ref', 'percpu', 'user' tag must be specified together with 'btf_id' tag */
+ if (ref_tag || percpu_tag || user_tag) {
ret = -EINVAL;
goto end;
}
return 0;
}
- /* referenced percpu btf_id pointer is not yet supported */
- if (ref_tag && percpu_tag) {
+ /* All three are mutually exclusive */
+ ret = ref_tag + percpu_tag + user_tag;
+ if (ret > 1) {
ret = -EINVAL;
goto end;
}
@@ -3257,6 +3265,8 @@ static int btf_find_field_kptr(const struct btf *btf, const struct btf_type *t,
flags |= BPF_MAP_VALUE_OFF_F_REF;
else if (percpu_tag)
flags |= BPF_MAP_VALUE_OFF_F_PERCPU;
+ else if (user_tag)
+ flags |= BPF_MAP_VALUE_OFF_F_USER;
tab->off[nr_off].offset = off;
tab->off[nr_off].btf_id = id;
--
2.35.1
Powered by blists - more mailing lists